﻿using System;
using System.Drawing;
using System.Windows.Forms;

namespace WindowsFormsApp1
{
    //Třída Character zajišťuje ovládání a pohyb postavy
    class Character
    {
        //Proměnné hráče, pozice v 2D prostoru, šířka hráče, rychlost, tlačítka, úhel, odkaz na herní mapu
        PointF position;
        public int Thick { get; private set; }
        public Point Position { get { return new Point((int)position.X,(int)position.Y); } }
        public Point DrawPos { get { return new Point((int)position.X - Thick / 2, (int)position.Y - Thick / 2); } }
        double maxSpeed = 2, acceleration = 0.2f, deceleration = 0.1f, xSpeed = 0, ySpeed = 0;
        bool u = false, d = false, l = false, r = false;
        public int angle = 0;
        Environment map;
        int panMid;
        int angleoffset = 0;

        //Konstruktor přijímá odkaz na herní mapu, střed zobrazovacího pole, a šířku hráče
        public Character(Environment map, int panMid, int thick)
        {
            this.map = map;
            this.panMid = panMid;
            Thick = thick;
            //Pozice hráče se při inicializaci nastaví na bod Spawn z objektu mapa, a vystředí podle šířky hráče
            position = new PointF(map.Spawn.X + thick / 2, map.Spawn.Y + thick / 2);
        }

        //Metoda KeyInput přijímá stisknuté tlačítko, a informaci o jeho stavu, true pro stlačení, false pro puštění
        public void KeyInput(Keys k, bool p)
        {
            //Switch přiřadí hodnotu stisknutí proměnné odpovídajícího směru
            switch(k)
            {
                case Keys.W: u = p; break;
                case Keys.S: d = p; break;
                case Keys.A: l = p; break;
                case Keys.D: r = p; break;
                default: break;
            }
            //Podle kombinace směrů je přiřazena hodnota odsazení směru pohybu
            if (u && l && !d && !r) angleoffset = -45;
            else if (u && r && !d && !l) angleoffset = 45;
            else if (l && !u && !d && !r) angleoffset = -90;
            else if (r && !u && !d && !l) angleoffset = 90;
            else if (d && l && !u && !r) angleoffset = -135;
            else if (d && r && !u && !l) angleoffset = 135;
            else if (d && !u && !l && !r) angleoffset = 180;
            else angleoffset = 0;
        }

        //Metoda FrameUpdate se odehraje při každém ticku GameTimeru, přijímá X souřadnici myši na herním panelu
        public int FrameUpdate(int mousex)
        {
            //K úhlu natočení hráče se přičte rozdíl X souřadnice myši a středu herního panelu, vydělený vhodnou konstantou pro úpravu citlivosti
            angle += (mousex - panMid) / 5;
            //Rychlost hráče po obou osách je přiblížena nule
            ySpeed = Math.Abs(ySpeed) < deceleration ? 0 : (ySpeed > 0 ? ySpeed - deceleration : (ySpeed < 0 ? ySpeed + deceleration : ySpeed + 0));
            xSpeed = Math.Abs(xSpeed) < deceleration ? 0 : (xSpeed > 0 ? xSpeed - deceleration : (xSpeed < 0 ? xSpeed + deceleration : xSpeed + 0));
            //Pokud je aktivní validní kombinace směrů (Né dva protichůdné) a pokud je rychlost po danné ose nižší než maximální rychlost,
            //rychlost se zvýší ve směru natočení hráče a odsazení směru pohybu
            if ((u ^ d) || (l ^ r)) 
            {
                ySpeed += Math.Abs(ySpeed) < maxSpeed ? Math.Sin((angle + angleoffset) * Math.PI / 180) * acceleration : 0; 
                xSpeed += Math.Abs(xSpeed) < maxSpeed ? Math.Cos((angle + angleoffset) * Math.PI / 180) * acceleration : 0; 
            }
            //Ze třídy mapa je zavolána metoda Collision která hráče na vhodném místě zastaví, pokud by se měl střetnout s překážkou
            map.Collision(maxSpeed, ref xSpeed, ref ySpeed, ref position);
            return angle;
        }

        //Metoda Package "zabalí" hodnoty směrových proměnných do jednoho čísla pomocí binárního kódování
        //pro snazší předání do třídy Replay
        public string Package()
        {
            return (u ? "1" : "0") + (d ? "1" : "0") + (l ? "1" : "0") + (r ? "1" : "0");
        }

        //Metoda pro řízení hráče záznamem
        public void Replay(string package)
        {
            string[] dirAng = package.Split(',');
            u = dirAng[0][0] == '1';
            d = dirAng[0][1] == '1';
            l = dirAng[0][2] == '1';
            r = dirAng[0][3] == '1';
            if (u && l && !d && !r) angleoffset = -45;
            else if (u && r && !d && !l) angleoffset = 45;
            else if (l && !u && !d && !r) angleoffset = -90;
            else if (r && !u && !d && !l) angleoffset = 90;
            else if (d && l && !u && !r) angleoffset = -135;
            else if (d && r && !u && !l) angleoffset = 135;
            else if (d && !u && !l && !r) angleoffset = 180;
            else angleoffset = 0;
            Int32.TryParse(dirAng[1], out angle);
            ySpeed = Math.Abs(ySpeed) < deceleration ? 0 : (ySpeed > 0 ? ySpeed - deceleration : (ySpeed < 0 ? ySpeed + deceleration : ySpeed + 0));
            xSpeed = Math.Abs(xSpeed) < deceleration ? 0 : (xSpeed > 0 ? xSpeed - deceleration : (xSpeed < 0 ? xSpeed + deceleration : xSpeed + 0));
            if ((u ^ d) || (l ^ r))
            {
                ySpeed += Math.Abs(ySpeed) < maxSpeed ? Math.Sin((angle + angleoffset) * Math.PI / 180) * acceleration : 0;
                xSpeed += Math.Abs(xSpeed) < maxSpeed ? Math.Cos((angle + angleoffset) * Math.PI / 180) * acceleration : 0;
            }
            map.Collision(maxSpeed, ref xSpeed, ref ySpeed, ref position);
        }
    }
}
