﻿using System;
using System.Drawing;
using System.Reflection;
using System.Windows.Forms;

namespace WindowsFormsApp1
{
    public partial class Form1 : Form
    {
        //Proměnné formuláře, objekt pro přístup k souboru, uživatelské rozhraní, herní prostředí, hráče, záznamu, proměnná pro zapnutí debugovacího
        //módu, počítadlo času a stav hry
        FileHandler fileHandler;
        UI userInterface;
        Environment map;
        Character player;
        Replay replay;
        bool debugMode = false;
        double time = 0;
        string gameState = "SizeSelect";
        string mapName = "";

        //Všem panelům je přiřazena vlastnost "DoubleBuffered" která zabraňuje problikávání při překreslování
        public Form1()
        {
            InitializeComponent();
            typeof(Panel).InvokeMember("DoubleBuffered", BindingFlags.SetProperty | BindingFlags.Instance | BindingFlags.NonPublic, null, MapPanel, new object[] { true });
            typeof(Panel).InvokeMember("DoubleBuffered", BindingFlags.SetProperty | BindingFlags.Instance | BindingFlags.NonPublic, null, CollisionPanel, new object[] { true });
            typeof(Panel).InvokeMember("DoubleBuffered", BindingFlags.SetProperty | BindingFlags.Instance | BindingFlags.NonPublic, null, MainView, new object[] { true });
        }

        //Při načtení formuláře se inicializuje objekt fileHandler a UI, pomocí kterých se zkontroluje celistvost souboru s herní mapou
        private void Form1_Load(object sender, EventArgs e)
        {
            userInterface = new UI(this.Size);
            replay = new Replay();
        }

        //Během chodu hry je aktivní GameTimer který každých 25 milisekund vyvolá tuto událost. Během ní se provede pohyb hráče, kurzor myši se
        //přemístí do středu obrazovky, aby nevyjela z formuláře, a aby bylo možné snímat o kolik se od středu pohnula. Dále je přičteno 0.025 k
        //proměnné která počítá kolik sekund hráči trvá vyhrát hru, překreslí se všechny panely, a zkontroluje se jestli hráč nevyhrál
        private void GameTimer_Tick(object sender, EventArgs e)
        {
            if (gameState == "GameOn")
            {
                replay.Record(player.Package(), player.FrameUpdate(MainView.PointToClient(MousePosition).X));
                Cursor.Position = new Point(this.Left + MainView.Width / 2, this.Top + MainView.Height / 2);
            }
            else if (gameState == "Replaying")
            {
                player.Replay(replay.Playback());
                if(replay.End)
                {
                    EndGame();
                    MessageBox.Show("Replay failed", "Error");
                }
            }
            time += 0.025f;
            MapPanel.Refresh();
            CollisionPanel.Refresh();
            MainView.Refresh();
            if (map.Victory) WinGame();
        }

        //Událost Form1_KeyDown odesílá hráči informaci o tom, která klávesa byla stisknuta. Dále hlídá speciální klávesy, Escape během chodu hry
        //vyvolá pauzu, a mezerník při zobrazení hlavního menu přepíná debugovací mód
        private void Form1_KeyDown(object sender, KeyEventArgs e)
        {
            if (e.KeyData == Keys.Escape && gameState == "GameOn") PauseGame();
            else if (e.KeyData == Keys.Escape && gameState == "Replaying") PauseReplay();
            else if (e.KeyData == Keys.Space && gameState == "MainMenu") DebugMode();
            if (gameState == "GameOn") player.KeyInput(e.KeyData, true);
        }

        //Událost Form1_KeyUp odesílá hráči informaci o tom, která klávesa byla puštěna
        private void Form1_KeyUp(object sender, KeyEventArgs e)
        {
            if (gameState == "GameOn") player.KeyInput(e.KeyData, false);
        }

        //Metoda Commander přebírá příkazy přijaté od klikání na uživatelské rozhraní, a na jejich základě aktivuje další metody
        private void Commander(string command)
        {
            switch(command)
            {
                case "500 x 250":
                    ResizeMyself(new Size(500, 250));
                    break;
                case "750 x 375":
                    ResizeMyself(new Size(750, 375));
                    break;
                case "1000 x 500":
                    ResizeMyself(new Size(1000, 500));
                    break;
                case "Ok":
                    gameState = "MainMenu";
                    MainView.Refresh();
                    break;
                case "Start":
                    replay.Clear();
                    InitializeGame();
                    break;
                case "Resume":
                    if (gameState == "Paused") StartGame();
                    else ResumeReplay();
                    break;
                case "Exit":
                    Application.Exit();
                    break;
                case "Editor":
                    EditMode();
                    break;
                case "Replay":
                    StartReplay();
                    break;
                case "Save":
                    SaveReplay();
                    break;
                default:
                    break;
            }
        }

        //MainView_Paint vykresluje na herní panel při každém jeho obnovení. Na základě herního stavu vykresluje buď některé menu z UI, nebo
        //herní pohled z první osoby
        private void MainView_Paint(object sender, PaintEventArgs e)
        {
            Graphics g = e.Graphics;
            switch (gameState)
            {
                case "SizeSelect":
                    g.DrawImage(userInterface.SizeSelectMenu(), new Point(0, 0));
                    break;
                case "MainMenu":
                    g.DrawImage(userInterface.MainMenu(), new Point(0, 0));
                    break;
                case "GameOn":
                    g.DrawImage(map.ThreeDee(player.Position, player.angle), new Rectangle(0,0,MainView.Width,MainView.Height));
                    break;
                case "Replaying":
                    g.DrawImage(map.ThreeDee(player.Position, player.angle), new Rectangle(0, 0, MainView.Width, MainView.Height));
                    break;
                case "Paused":
                    g.DrawImage(map.ThreeDee(player.Position, player.angle), new Point(0, 0));
                    g.DrawImage(userInterface.PauseMenu(), new Point(0, 0));
                    break;
                case "PausedReplay":
                    g.DrawImage(map.ThreeDee(player.Position, player.angle), new Point(0, 0));
                    g.DrawImage(userInterface.PauseMenu(), new Point(0, 0));
                    break;
                case "GameEnd":
                    g.DrawImage(userInterface.WinMenu((int)time), new Point(0, 0));
                    break;
                default:
                    break;
            }
        }

        //Při kliknutí na herní panel je kliknutí odesláno UI které vyhodnotí jestli se jedná o kliknutí na nějaké tlačítko, a pošle jeho příkaz
        //zpátky metodě Commander
        private void MainView_MouseClick(object sender, MouseEventArgs e)
        {
            Commander(userInterface.UIClick(MainView.PointToClient(MousePosition)));
        }

        //Metoda ResizeMyself mění velikost formuláře, herního panelu, UI a přemisťuje formulář do středu obrazovky
        private void ResizeMyself(Size newSize)
        {
            this.Size = newSize;
            MainView.Size = newSize;
            userInterface.Resize(newSize);
            this.CenterToScreen();
            MainView.Refresh();
        }

        //Metoda InitializeGame inicializuje herní prostředí a hráče pomocí informací získaných třídou fileHandler. Pokud už bylo herní prostředí
        //inicializováno, pouze vyvolá jeho metodu reset. Také posune sekundární panely na okraj hlavního herního panelu
        private void InitializeGame()
        {
            //if (map == null) map = new Environment(MainView.Size, fileHandler.Map, fileHandler.CollisionMap, fileHandler.Functions, fileHandler.Textures, fileHandler.Spawn);
            //else map.Reset();
            OpenFileDialog ofd = new OpenFileDialog();
            ofd.Filter = "png|*.png";
            if(ofd.ShowDialog() == DialogResult.OK)
            {
                mapName = ofd.FileName;
                fileHandler = new FileHandler(mapName);
                if (userInterface.WeHappy(fileHandler.Error))
                {
                    map = new Environment(MainView.Size, fileHandler.Map, fileHandler.CollisionMap, fileHandler.Functions, fileHandler.Textures, fileHandler.Spawn);
                    player = new Character(map, MainView.Width / 2, fileHandler.Thick);
                    MapPanel.Left = MainView.Width;
                    CollisionPanel.Left = MainView.Width;
                    Cursor.Hide();
                    GameTimer.Enabled = true;
                    gameState = "GameOn";
                    Cursor.Position = new Point(this.Left + MainView.Width / 2, this.Top + MainView.Height / 2);
                }
            }
        }

        //Metoda StartGame skryje kurzor myši, aktivuje GameTimer, a aktivuje stav "ve hře"
        private void StartGame()
        {
            Cursor.Hide();
            GameTimer.Enabled = true;
            gameState = "GameOn";
        }

        //Metoda PauseGame odkryje kurzor myši, zastaví GameTimer, aktivuje stav "zastaveno" a překreslí herní panel pro zobrazení UI menu
        private void PauseGame()
        {
            Cursor.Show();
            GameTimer.Enabled = false;
            gameState = "Paused";
            MainView.Refresh();
        }

        //Metoda pro pozastavení záznamu
        private void PauseReplay()
        {
            Cursor.Show();
            GameTimer.Enabled = false;
            gameState = "PausedReplay";
            MainView.Refresh();
        }

        //Metoda pro obnovení přehrávání záznamu
        private void ResumeReplay()
        {
            Cursor.Hide();
            GameTimer.Enabled = true;
            gameState = "Replaying";
        }

        //Metoda pro uložení záznamu
        private void SaveReplay()
        {
            SaveFileDialog sfd = new SaveFileDialog();
            sfd.Filter = "txt|*.txt";
            if (sfd.ShowDialog() == DialogResult.OK)
            {
                replay.Save(sfd.FileName, mapName);
                gameState = "MainMenu";
            }
        }

        //Metoda WinGame odkryje kurzor myši, zastaví GameTimer, aktivuje stav "konec hry" vykreslí výherní obrazovku z UI, a resetuje hru aby bylo
        //možno začít znovu
        private void WinGame()
        {
            Cursor.Show();
            GameTimer.Enabled = false;
            gameState = "GameEnd";
            MainView.Refresh();
            time = 0;
        }

        //Metoda pro ukončení hry bez výhry
        private void EndGame()
        {
            Cursor.Show();
            GameTimer.Enabled = false;
            gameState = "MainMenu";
            MainView.Refresh();
            time = 0;
        }

        //Metoda pro spuštění záznamu
        private void StartReplay()
        {
            OpenFileDialog ofd = new OpenFileDialog()
            {
                Filter = "txt|*.txt"
            };
            if(ofd.ShowDialog() == DialogResult.OK)
            {
                mapName = replay.Load(ofd.FileName);
                fileHandler = new FileHandler(mapName);
                if (userInterface.WeHappy(fileHandler.Error))
                {
                    map = new Environment(MainView.Size, fileHandler.Map, fileHandler.CollisionMap, fileHandler.Functions, fileHandler.Textures, fileHandler.Spawn);
                    player = new Character(map, MainView.Width / 2, fileHandler.Thick);
                    MapPanel.Left = MainView.Width;
                    CollisionPanel.Left = MainView.Width;
                    Cursor.Hide();
                    GameTimer.Enabled = true;
                    gameState = "Replaying";
                    Cursor.Position = new Point(this.Left + MainView.Width / 2, this.Top + MainView.Height / 2);
                }
            }
        }

        //Metoda pro spuštění editoru
        private void EditMode()
        {
            Editor editor = new Editor();
            editor.ShowDialog();
        }

        //Metoda DebugMode přepíná proměnnou debugMode, a na jejím základě buď roztahuje nebo smršťuje formulář. Tímto zobrazí dva skryté panely
        //které jsou za normálních podmínek mimo formulář.
        private void DebugMode()
        {
            if (debugMode)
            {
                this.Size = MainView.Size;
            }
            else
            {
                this.Width = MainView.Width + 250;
                this.Height = 500;
            }
            debugMode = !debugMode;
        }

        //Událost MapPanel_Paint vykresluje na vedlejší formulář grafickou mapu, pozici, a směr hráče, pokud je aktivní debugMode
        private void MapPanel_Paint(object sender, PaintEventArgs e)
        {
            if (debugMode)
            {
                Graphics g = e.Graphics;
                g.DrawImage(map.Map, new Point(0, 0));
                g.FillEllipse(Brushes.DarkBlue, new Rectangle(player.DrawPos, new Size(player.Thick, player.Thick)));
                g.DrawLine(new Pen(Color.Aqua, 3), player.Position, new Point((int)(player.Position.X + Math.Cos(player.angle * Math.PI / 180) * map.ScanDistance), (int)(player.Position.Y + Math.Sin(player.angle * Math.PI / 180) * map.ScanDistance))); ;
            }
        }

        //Událost MapPanel_Paint vykresluje na vedlejší formulář kolizní mapu, pozici, a směr hráče, pokud je aktivní debugMode
        private void CollisionPanel_Paint(object sender, PaintEventArgs e)
        {
            if(debugMode)
            {
                Graphics g = e.Graphics;
                g.DrawImage(map.ColisionMap, new Point(0, 0));
                g.FillEllipse(Brushes.DarkBlue, new Rectangle(player.DrawPos, new Size(player.Thick, player.Thick)));
                g.DrawLine(new Pen(Color.Aqua, 3), player.Position, new Point((int)(player.Position.X + Math.Cos(player.angle * Math.PI / 180) * map.ScanDistance), (int)(player.Position.Y + Math.Sin(player.angle * Math.PI / 180) * map.ScanDistance))); ;
            }
        }
    }
}
